#version 330

const float VIGNETTE_INTENSITY = 15.0;
const float VIGNETTE_RANGE = 0.15;
const float DEBUG_VIEW_RENDER_WIDTH = 0.2f;
const float DEBUG_VIEW_RENDER_HEIGHT = 0.2f;
const float PI = 3.14159265;
//---------------------------------------------------------------------------
in vec2 TexCoords;
out vec4 outColor;
//---------------------------------------------------------------------------
uniform sampler2D gScreenTexture;
uniform sampler2D gBloomTexture;
uniform sampler2D gDepthTexture;
uniform sampler2D gLensFlareTexture;
uniform float gExposure;
uniform int gToneMappingSwitch;
uniform bool gBloom;
uniform bool gDOFSwitch;
uniform bool gVignetteSwitch;
uniform sampler2D gDebugTexture;
uniform bool gDebugView;
uniform bool gDebugRenderDOFArea;
//---------------------------------------------------------------------------
float A = 0.15;
float B = 0.50;
float C = 0.10;
float D = 0.20;
float E = 0.02;
float F = 0.30;
float W = 11.2;
vec3 Uncharted2Tonemap(vec3 x)
{
   return ((x*(A*x+C*B)+D*E)/(x*(A*x+B)+D*F))-E/F;
}
//---------------------------------------------------------------------------
vec2 m_TexelSize = vec2(1.0/1600, 1.0/900);
uniform float gFocalPoint = 0.5;  
uniform float gFocalLength = 0.1;
uniform float gFocalFade = 0.1;
uniform int gDOFSamples = 3; //samples on the first ring
uniform int gDOFRings = 3; //ring count
uniform float gFocalMaxBlur = 1.0; //clamp value of max blur (0.0 = no blur,1.0 default)
uniform float gDOFBias = 0.5; //bokeh edge bias
uniform float gDOFFringe = 0.7; //bokeh chromatic aberration/fringing
//---------------------------------------------------------------------------
vec3 GetColorBlurredAndFringed(vec2 a_coords,float a_blur) //processing the sample
{
	// sample diffuse texture + fringe and blur
	// fringe creates a coloroffset
	vec3 color = vec3(0);
	color.r = texture(gScreenTexture, a_coords + vec2(0.000	,  1.0) * m_TexelSize * gDOFFringe * a_blur).r;
	color.g = texture(gScreenTexture, a_coords + vec2(-0.866	, -0.5) * m_TexelSize * gDOFFringe * a_blur).g;
	color.b = texture(gScreenTexture, a_coords + vec2(0.866	, -0.5) * m_TexelSize * gDOFFringe * a_blur).b;

	return color;
}
//---------------------------------------------------------------------------
float GetDOFBlurFactor(vec4 a_depth)
{
	if (a_depth.g >= 1)
	{
		return 0;
	}

	float depthDiff = abs(a_depth.r - gFocalPoint);
	float blur = (depthDiff - gFocalLength) / gFocalFade;
	return clamp(blur, 0.0, 1.0);
}
//---------------------------------------------------------------------------
vec3 ProcessDOF(vec3 color)
{
	vec2 texsize = textureSize(gScreenTexture, 0);
	m_TexelSize = vec2(1.0/texsize.x, 1.0/texsize.y);

	vec4 pixelDepth = texture(gDepthTexture, TexCoords);

	//dof blur factor calculation
	float currentPixelBlur = GetDOFBlurFactor(pixelDepth);

	// debug render
	if (gDebugRenderDOFArea)
	{
		return vec3(1 - currentPixelBlur);
	}
	
	// getting blur x and y step factor
	float wStep = m_TexelSize.x * currentPixelBlur * gFocalMaxBlur;
	float hStep = m_TexelSize.y * currentPixelBlur * gFocalMaxBlur;

	float sampleCount = 1.0;
	int ringsamples;

	for (int i = 1; i <= gDOFRings; i++)
	{   
		ringsamples = i * gDOFSamples;
			
		for (int j = 0 ; j < ringsamples ; j++)   
		{
			float stepAmount = PI * 2.0 / float(ringsamples);
			float pw = (cos(float(j) * stepAmount) * float(i));
			float ph = (sin(float(j) * stepAmount) * float(i));

			// get coordinates to sample from
			vec2 coordinates = TexCoords.xy + vec2(pw * wStep, ph * hStep);

			// check blur difference at that sample with the original (if its much too great, dont sample it will create halos)
			vec4 depthAtSample = texture(gDepthTexture, coordinates);
			float depthDiff = abs(currentPixelBlur - GetDOFBlurFactor(depthAtSample));

			if (depthDiff < 0.3)
			{
				// blur and fringe the pixel
				vec3 colorBlurredFringed = GetColorBlurredAndFringed(coordinates, currentPixelBlur);

				float ringFactor = mix(1.0, (float(i)) / (float(gDOFRings)), gDOFBias);  

				color += colorBlurredFringed * ringFactor;

				sampleCount += ringFactor;
			}
		}
	}
	color /= sampleCount;

	return color;
}
//---------------------------------------------------------------------------
void main()
{ 
    vec3 hdrColor = texture(gScreenTexture, TexCoords).rgb;

	// dof
	if (gDOFSwitch)
	{
		hdrColor = ProcessDOF(hdrColor);
	}

	// lens flare
	hdrColor += texture(gLensFlareTexture, TexCoords).rgb;

	// bloom  
    if(gBloom) 
	{
	    vec3 bloomColor = texture(gBloomTexture, TexCoords).rgb;
		hdrColor += bloomColor;
	}

	// exposure
	vec3 mapped = hdrColor * gExposure;

	// tone mapping
    const float gamma = 2.2;
	int toneMapType = gToneMappingSwitch % 4;
	if (toneMapType == 0)
	{
		// 0 - linear
	    mapped = pow(mapped, vec3(1.0 / gamma));
	}
	else if (toneMapType == 1)
	{
		// 1 - Reinhard
		mapped = mapped/(1+mapped);
	    mapped = pow(mapped, vec3(1.0 / gamma));
	}
	else if (toneMapType == 2)
	{
		// 2 - Filmic   
		vec3 x = max(vec3(0), vec3(mapped-0.004));
		mapped = (x*(6.2*x+.5))/(x*(6.2*x+1.7)+0.06);
	}
	else if (toneMapType == 3)
	{
		// 3 - Uncharted 2
		float ExposureBias = 2.0f;
		vec3 curr = Uncharted2Tonemap(ExposureBias*mapped);

		vec3 whiteScale = 1.0f/Uncharted2Tonemap(vec3(W));
		mapped = curr*whiteScale;
	    mapped = pow(mapped, vec3(1.0 / gamma));
	}
  
	// vignette
	if (gVignetteSwitch)
	{
		vec2 uv = TexCoords.xy;
		uv *=  1.0 - uv.yx;
		float vig = uv.x*uv.y * VIGNETTE_INTENSITY;
		vig = pow(vig, VIGNETTE_RANGE); // change pow for modifying the extend of the  vignette
    
		mapped = mapped * vig;
	}

	outColor = vec4(mapped, 1.0);

	// debug render
	if (gDebugView && TexCoords.x < DEBUG_VIEW_RENDER_WIDTH && TexCoords.y < DEBUG_VIEW_RENDER_HEIGHT)
	{
		outColor = texture(gDebugTexture, vec2(TexCoords.x / DEBUG_VIEW_RENDER_WIDTH, TexCoords.y / DEBUG_VIEW_RENDER_HEIGHT));
	}	
}
