#version 330
                                               
const int MAX_POINT_LIGHTS = 2;     

in vec2 TexCoord;
in vec3 Normal;
in vec3 WorldPos;

out vec4 outColor;

struct BaseLight
{
    vec3  Color;
    float AmbientIntensity;
	float DiffuseIntensity;
};

struct DirectionalLight
{
	BaseLight Base;
	vec3 Direction;
};

struct Attenuation
{
    float Constant;
    float Linear;
    float Exp;
};

struct PointLight
{
    BaseLight Base;
    vec3 Position;
    Attenuation Atten;
};

uniform DirectionalLight gDirLight;
uniform sampler2D gSampler;
uniform vec3 gCameraPos;                                                      
uniform float gMatSpecularIntensity;
uniform float gSpecularPower;
                                                         
uniform int gNumPointLights;                                                                
uniform PointLight gPointLights[MAX_POINT_LIGHTS];                                          
                                                                                            
vec4 CalcLightInternal(BaseLight _Light, vec3 _LightDirection, vec3 _Normal)                   
{
    vec4 AmbientColor = vec4(_Light.Color * _Light.AmbientIntensity, 1.0f);
    float DiffuseFactor = dot(_Normal, -_LightDirection);                                     
                                                                                            
    vec4 DiffuseColor  = vec4(0, 0, 0, 0);                                                  
    vec4 SpecularColor = vec4(0, 0, 0, 0);                                                  
                                                                                            
    if (DiffuseFactor > 0) 
	{                                                                
        DiffuseColor = vec4(_Light.Color * _Light.DiffuseIntensity * DiffuseFactor, 1.0f);    
                                                                                            
        vec3 VertexToEye = normalize(gCameraPos - WorldPos);                             
        vec3 LightReflect = normalize(reflect(_LightDirection, _Normal));                     
        float SpecularFactor = dot(VertexToEye, LightReflect);                              
        if (SpecularFactor > 0) 
		{                                                           
            SpecularFactor = pow(SpecularFactor, gSpecularPower);
            SpecularColor = vec4(_Light.Color * gMatSpecularIntensity * SpecularFactor, 1.0f);                         
        }                                                                                   
    }

    return (AmbientColor + DiffuseColor + SpecularColor);
}
                                                                                            
vec4 CalcDirectionalLight(vec3 normal)                                                      
{
    return CalcLightInternal(gDirLight.Base, gDirLight.Direction, normal);  
}                                                                                               

vec4 CalcPointLight(PointLight l, vec3 normal)
{
    vec3 LightDirection = WorldPos - l.Position;
    float Distance = length(LightDirection);
    LightDirection = normalize(LightDirection);

    vec4 Color = CalcLightInternal(l.Base, LightDirection, normal);
    float Attenuation =  l.Atten.Constant + 
                         l.Atten.Linear * Distance +
                         l.Atten.Exp * Distance * Distance;

    return Color / Attenuation;
}

void main()
{
	vec3 normal = normalize(Normal);
	vec4 light = CalcDirectionalLight(normal);
	
	for (int i = 0 ; i < gNumPointLights ; i++) 
	{
        light += CalcPointLight(gPointLights[i], normal);
    }

	outColor = texture(gSampler, TexCoord) * light;
}