#version 330

#define EPSILON -0.0000
#define USE_SAMPLER2DSHADOW
                
const int MAX_POINT_LIGHTS = 3;
const int NUM_CASCADES = 4;

vec4 m_DebugCascadeIndicator;

vec3 SAMPLE_OFFSET_DIRECTIONS[20] = vec3[]
(
   vec3( 1,  1,  1), vec3( 1, -1,  1), vec3(-1, -1,  1), vec3(-1,  1,  1), 
   vec3( 1,  1, -1), vec3( 1, -1, -1), vec3(-1, -1, -1), vec3(-1,  1, -1),
   vec3( 1,  1,  0), vec3( 1, -1,  0), vec3(-1, -1,  0), vec3(-1,  1,  0),
   vec3( 1,  0,  1), vec3(-1,  0,  1), vec3( 1,  0, -1), vec3(-1,  0, -1),
   vec3( 0,  1,  1), vec3( 0, -1,  1), vec3( 0, -1, -1), vec3( 0,  1, -1)
); 

struct BaseLight
{
    vec3  Color;
    float AmbientIntensity;
	float DiffuseIntensity;
};

struct DirectionalLight
{
	BaseLight Base;
	vec3 Direction;
};

struct Attenuation
{
    float Constant;
    float Linear;
    float Exp;
};

struct PointLight
{
    BaseLight Base;
    vec3 Position;
    Attenuation Atten;
	bool ShadowMapEnabled;
	samplerCube ShadowMap;
};
//---------------------------------------------------------------------------
uniform sampler2D gSampler;
uniform sampler2D gNormalMap;
uniform bool gNormalMapEnabled;
uniform sampler2D gSpecularMap;
uniform bool gSpecularMapEnabled;
uniform sampler2D gGlowMap;
uniform bool gGlowMapEnabled;
#ifdef USE_SAMPLER2DSHADOW
uniform sampler2DShadow gShadowMapDir[NUM_CASCADES];
#else
uniform sampler2D gShadowMapDir[NUM_CASCADES];
#endif
uniform float gShadowMapCascadeEndClipSpace[NUM_CASCADES];
uniform float gSpecularPower;
uniform vec2 gShadowMapSize;
uniform vec3 gCameraPos;
uniform int gNumPointLights;
uniform DirectionalLight gDirLight;
uniform PointLight gPointLights[MAX_POINT_LIGHTS];
uniform bool gDebugShowCSM;
//---------------------------------------------------------------------------
in vec4 s_LightSpacePos[NUM_CASCADES];
in float s_ClipSpacePosZ;
in vec2 s_TexCoord;
in vec3 s_Normal;
in vec3 s_CameraPos;
in vec3 s_WorldPos;
in vec3 s_Pos;
in vec3 s_DirLightDir;
in vec3 s_PointLightsPos[MAX_POINT_LIGHTS];
//---------------------------------------------------------------------------
layout (location = 0) out vec4 outColor;
layout (location = 1) out vec4 outBrightColor;
layout (location = 2) out vec4 outDepth;
//---------------------------------------------------------------------------
float CalcShadowFactorPoint(PointLight l)
{
	if (!l.ShadowMapEnabled)
	{
		return 1.0;
	}

	vec3 lightDir = s_WorldPos - l.Position;
	float lightDist = length(lightDir);
	float shadow = 0.0;
	float bias = 0.15;
	float viewDist = length(gCameraPos - s_WorldPos);
	float diskRadius = (1.0 + normalize(viewDist)) / 25.0;  

	for(int i = 0; i < 20; i++)
	{
		float closestDepth = texture(l.ShadowMap, lightDir + SAMPLE_OFFSET_DIRECTIONS[i] * diskRadius).r;
		if(lightDist - bias > closestDepth)
		{
			shadow += 1.0;
		}
	}
	shadow /= 20.0;  
	
	return 1.0f - shadow;
}
//---------------------------------------------------------------------------
float CalcShadowFactorDir(int cascadeIndex, vec4 lightSpacePos)
{
	vec3 projCoords = lightSpacePos.xyz / lightSpacePos.w;
	vec2 uvs;
	uvs.x = 0.5 * projCoords.x + 0.5;
	uvs.y = 0.5 * projCoords.y + 0.5;
	float z = 0.5 * projCoords.z + 0.5;

	float shadow = 0.0;
	float xOffset = 1.0 / gShadowMapSize.x;
	float yOffset = 1.0 / gShadowMapSize.y;
	for(int x = -1; x <= 1; ++x)
	{
		for(int y = -1; y <= 1; ++y)
		{
			vec2 offsets = vec2(x * xOffset, y * yOffset);
			vec2 sampleUV = uvs.xy + offsets;
			if (sampleUV.x < 0.0 || sampleUV.x > 1.0 || sampleUV.y < 0.0 || sampleUV.y > 1.0)
			{
				shadow += 0.0;
			}
			else
			{
#ifdef USE_SAMPLER2DSHADOW
				float pcfDepth = texture(gShadowMapDir[cascadeIndex], vec3(sampleUV, z));
				shadow += pcfDepth;
#else
				float pcfDepth = texture(gShadowMapDir[cascadeIndex], sampleUV).r; 
				shadow += (pcfDepth < z + EPSILON) ? 0.0 : 1.0;        
#endif
			}            
		}    
	}

	shadow /= 9.0;
	return shadow;
}
//---------------------------------------------------------------------------
vec4 CalcLightInternal(BaseLight light, vec3 lightDir, vec3 normal, float shadowFactor)
{
    vec4 ambientCol = vec4(light.Color * light.AmbientIntensity, 1.0f);
    vec4 diffuseCol  = vec4(0, 0, 0, 0);
    vec4 specularCol = vec4(0, 0, 0, 0);

    float diffuseFactor = dot(normal, -lightDir);
    if (diffuseFactor > 0) 
	{
        diffuseCol = vec4(light.Color * light.DiffuseIntensity * diffuseFactor, 1.0f);
		
		vec3 vertToCamera = normalize(s_CameraPos - s_Pos);
        vec3 lightReflect = normalize( reflect(lightDir, normal) );
        float specularFactor = dot(vertToCamera, lightReflect);
        if (specularFactor > 0)
		{
            specularFactor = pow(specularFactor, gSpecularPower);

			if (gSpecularMapEnabled)
			{
	            specularCol = vec4(light.Color * texture(gSpecularMap, s_TexCoord).rgb * specularFactor, 1.0f);
			}
			else
			{
		        specularCol = vec4(light.Color * specularFactor, 1.0f);
			}
        }
    }

    return ambientCol + shadowFactor * (diffuseCol + specularCol);
}                                                                                           
//---------------------------------------------------------------------------
vec4 CalcDirectionalLight(vec3 normal)
{
	// CSM
	float shadowFactor = 0.0;
    m_DebugCascadeIndicator = vec4(0.0, 0.0, 0.0, 0.0);

    for (int i = 0 ; i < NUM_CASCADES ; i++) 
	{
        if (s_ClipSpacePosZ <= gShadowMapCascadeEndClipSpace[i]) 
		{
			shadowFactor = CalcShadowFactorDir(i, s_LightSpacePos[i]);

			if (gDebugShowCSM)
			{
				m_DebugCascadeIndicator = i % 4 == 0 ? vec4(0.05, 0.05, 0.0, 0.0) : 
										  i % 4 == 1 ? vec4(0.0, 0.05, 0.05, 0.0) : 
										  i % 4 == 2 ? vec4(0.05, 0.0, 0.05, 0.0) : 
													   vec4(0.1, 0.00, 0.0, 0.0);
			}
            break;
        }
   }
       
	// calculate light
	vec3 lightDir = normalize(s_DirLightDir);
	return CalcLightInternal(gDirLight.Base, lightDir, normal, shadowFactor);
}
//---------------------------------------------------------------------------
vec4 CalcPointLight(PointLight l, vec3 lightPos, vec3 normal)
{
	// calculate shadow
	float shadowFactor = CalcShadowFactorPoint(l);
	
	// calculate light
    vec3 lightDir = s_Pos - lightPos;
    float lightDist = length(lightDir);
    lightDir = normalize(lightDir);

    vec4 col = CalcLightInternal(l.Base, lightDir, normal, shadowFactor);
    float atten = l.Atten.Constant + 
				  l.Atten.Linear * lightDist +
				  l.Atten.Exp * lightDist * lightDist;

    return col / atten;
} 
//---------------------------------------------------------------------------
float near = 0.5; 
float LinearizeDepth(float depth) 
{
	float far = gShadowMapCascadeEndClipSpace[NUM_CASCADES-1];
    float z = depth * 2.0 - 1.0; // Back to NDC 
    return (2.0 * near * far) / (far + near - z * (far - near));	
}
//---------------------------------------------------------------------------
void main()
{
	// Sample texture
	outColor = texture(gSampler, s_TexCoord);

	// cutout alpha
	if (outColor.a < 0.5)
	{
		discard; 
	}

	// Normal Maps
	vec3 normal = vec3(0);
	if( gNormalMapEnabled )
	{
		// Obtain normal from normal map in range [0,1]
		normal = texture(gNormalMap, s_TexCoord).rgb;
		normal = normalize(normal * 2.0 - 1.0);   
	}
	else
	{
		normal = normalize(s_Normal);
	}

	// Light pass
    vec4 light = CalcDirectionalLight(normal);
	for (int i = 0 ; i < gNumPointLights ; i++) 
    {
        light += CalcPointLight(gPointLights[i], s_PointLightsPos[i], normal);
    }
	
	outColor = outColor * light;

    // Sample bloom
    float brightness = dot(outColor.rgb, vec3(0.2126, 0.7152, 0.0722));
    if(brightness > 1.0)
	{
        outBrightColor = vec4(outColor.rgb, 1.0);
	}
	else
	{
		outBrightColor = vec4(0.0f, 0.0f, 0.0f, 1.0f);
	}

	// glow map
	if (gGlowMapEnabled)
	{
		vec3 glow = vec3(texture(gGlowMap, s_TexCoord).rgb) * 2.0f;
		outBrightColor += vec4(glow, 1.0);
	}

	float far = gShadowMapCascadeEndClipSpace[NUM_CASCADES-1];
    float depth = LinearizeDepth(gl_FragCoord.z) / far;
    outDepth = vec4(depth, 0.0, 0.0, 1.0);

	// debug
	if (gDebugShowCSM)
	{
		outColor += m_DebugCascadeIndicator;
	}
}
